#define gc (vm->gc)

void tp_grey(TP,tp_obj v) {
    int n;
    if (v.type == TP_NONE || v.type == TP_NUMBER) { return; }
    if ((v.type == TP_FNC || v.type == TP_DEF) &&
        (v.fnc.self == 0 && v.fnc.globals == 0)) { return; }
    if ((n=_tp_dict_find(vm,gc->white,v)) == -1) { return; }

    // must grab the actual key, because sometimes the value 
    // can be different (strings are equal to strings but they aren't
    // always equal if they are from bytecode.
    v = gc->white->items[n].key;
    _tp_dict_del(vm,gc->white,v,"tp_grey");
    if (v.type == TP_STRING || v.type == TP_DATA) {
        _tp_list_appendx(vm,gc->black,v);
        return;
    }
    _tp_list_appendx(vm,gc->grey,v);
}

void tp_follow(TP,tp_obj v) {
    int type = v.type;
    if (type == TP_LIST || type == TP_PARAMS) {
        int n;
        for (n=0; n<v.list.val->len; n++) {
            tp_grey(vm,v.list.val->items[n]);
        }
    }
    if (type == TP_DICT || type == TP_PARAMS) {
        int i;
        for (i=0; i<v.dict.val->len; i++) {
            int n = _tp_dict_next(vm,v.dict.val);
            tp_grey(vm,v.dict.val->items[n].key);
            tp_grey(vm,v.dict.val->items[n].val);
        }
    }
    if (type == TP_FNC || type == TP_DEF) {
        if (v.fnc.self) { tp_grey(vm,*(tp_obj*)v.fnc.self); }
        if (type == TP_DEF && v.def.globals) {
            tp_grey(vm,*(tp_obj*)v.def.globals);
        }
    }
}

void tp_reset(TP) {
    while (gc->black->len) {
        tp_obj v = _tp_list_pop(vm,gc->black,gc->black->len-1,"tp_reset");
        _tp_dict_setx(vm,gc->white,v,True);
    }
}

void tp_gc_init(TP) {
    tp_gc *v = tp_malloc(sizeof(tp_gc));
    v->white = _tp_dict_new();
    v->strings = _tp_dict_new();
    v->grey = _tp_list_new(0);
    v->black = _tp_list_new(0);
    v->target = 256;
    v->steps = 0;
    gc = v;
}

void tp_gc_deinit(TP) {
    _tp_dict_free(gc->white);
    _tp_dict_free(gc->strings);
    _tp_list_free(gc->grey);
    _tp_list_free(gc->black);
    tp_free(gc);
}

void tp_delete(TP,tp_obj v) {
    int type = v.type;
    if (type == TP_LIST) {
        _tp_list_free(v.list.val);
        return;
    } else if (type == TP_DICT) {
        _tp_dict_free(v.dict.val);
        return;
    } else if (type == TP_PARAMS) {
        _tp_list_free(v.list.val);
        _tp_dict_free(v.dict.val);
        return;
    } else if (type == TP_STRING) {
        tp_free(v.string.val);
        return;
    } else if (type == TP_DATA) {
        if (v.data.meta && v.data.meta->free) {
            v.data.meta->free(vm,v);
        }
        return;
    } else if (type == TP_FNC || type == TP_DEF) {
        if (v.def.self) { tp_free(v.def.self); }
        if (type == TP_DEF && v.def.globals) { tp_free(v.def.globals); }
        return;
    }
    tp_raise(,"tp_delete(%s)",STR(v));
}

void tp_collect(TP) {
    int i;
    for (i=0; i<gc->white->len; i++) {
        int n = _tp_dict_next(vm,gc->white);
        tp_obj r = gc->white->items[n].key;
        if (r.type == TP_STRING) {
            //this can't be moved into tp_delete, because tp_delete is
            // also used by tp_track_s to delete redundant strings
            _tp_dict_del(vm,gc->strings,r,"tp_collect");
        }
        tp_delete(vm,r);
    }
    _tp_dict_reset(gc->white);
    tp_reset(vm);
}

void _tp_gcinc(TP) {
    if (!gc->grey->len) { return; }
    tp_obj v = _tp_list_pop(vm,gc->grey,gc->grey->len-1,"_tp_gcinc");
    tp_follow(vm,v);
    _tp_list_appendx(vm,gc->black,v);
}

void tp_full(TP) {
    while (gc->grey->len) {
        _tp_gcinc(vm);
    }
    tp_collect(vm);
    tp_follow(vm,vm->root);
}

void tp_gcinc(TP) {
    gc->steps += 1;
    if (gc->steps < TP_GCMAX || gc->grey->len > 0) {
        _tp_gcinc(vm); _tp_gcinc(vm);
    }
    if (gc->steps < TP_GCMAX || gc->grey->len > 0) { return; }
    gc->steps = 0;
    tp_full(vm);
    return;
}

tp_obj tp_track_s(TP,tp_obj v) {
    int i = _tp_dict_find(vm,gc->strings,v);
    if (i != -1) {
//         fprintf(stderr,"tp_track_s: %d/%d %s %s\n",i,gc->strings->len,STR(v),STR(gc->strings->items[i].key));
        tp_delete(vm,v); return gc->strings->items[i].key; 
    }
    _tp_dict_setx(vm,gc->strings,v,True);
    tp_track(vm,v);
    return v;
}

void tp_track(TP,tp_obj v) {
    if (v.type == TP_NONE || v.type == TP_NUMBER) { return; }
    if ((v.type == TP_FNC || v.type == TP_DEF) &&
        (v.fnc.self == 0 && v.fnc.globals == 0)) { return; }
    tp_gcinc(vm);
    if (v.type == TP_STRING || v.type == TP_DATA) {
        _tp_list_appendx(vm,gc->black,v);
        return;
    }
//     _tp_list_appendx(vm,gc->black,v);
    _tp_list_appendx(vm,gc->grey,v);
}

//

