#include "tp.h"
#include "list.c"
#include "dict.c"
#include "util.c"
#include "string.c"
#include "builtins.c"
#include "tgc.c"


tp_obj tp_str(TP,tp_obj self) {
    int type = self.type;
    if (type == TP_STRING) { return self; }
    if (type == TP_NUMBER) {
        return _tp_number2char(vm,self.number.val);
    } else if(type == TP_DICT) {
        return tp_printf(vm,"<dict 0x%x>",self.dict.val);
    } else if(type == TP_LIST) {
        return tp_printf(vm,"<list 0x%x>",self.list.val);
    } else if(type == TP_PARAMS) {
        return tp_printf(vm,"<params 0x%x>",self.list.val);
    } else if (type == TP_NONE) {
        return tp_string("None");
    } else if (type == TP_DATA) {
        return tp_printf(vm,"<data 0x%x>",self.data.val);
    } else if (type == TP_FNC) {
        return tp_printf(vm,"<fnc 0x%x>",self.fnc.val);
    } else if (type == TP_DEF) {
        return tp_printf(vm,"<def 0x%x>",self.def.val);
    }
    return tp_string("<?>");
}


tp_obj tp_has(TP,tp_obj self, tp_obj k) {
    int type = self.type;
    if (type == TP_PARAMS && self.list.val->len > 0) { return True; }
    if (type == TP_DICT || type == TP_PARAMS) {
        if (_tp_dict_find(vm,self.dict.val,k) != -1) { return True; }
        return False;
    } else if (type == TP_STRING && k.type == TP_STRING) {
        char *p = strstr(tp_cstring(self),tp_cstring(k));
        return tp_number(p != 0);
    } else if (type == TP_LIST) {
        return tp_number(_tp_list_find(vm,self.list.val,k)!=-1);
    }
    tp_raise(None,"tp_has(%s,%s)",STR(self),STR(k));
}

tp_obj tp_del(TP,tp_obj self, tp_obj k) {
    int type = self.type;
    if (type == TP_DICT) {
        _tp_dict_del(vm,self.dict.val,k,"tp_del");
        return None;
    }
    tp_raise(None,"tp_del(%s,%s)",STR(self),STR(k));
}


tp_obj tp_iter(TP,tp_obj self, tp_obj k) {
    int type = self.type;
    if (type == TP_LIST || type == TP_STRING) { return tp_get(vm,self,k); }
    if (type == TP_DICT && k.type == TP_NUMBER) {
        return self.dict.val->items[_tp_dict_next(vm,self.dict.val)].key;
    }
    tp_raise(None,"tp_iter(%s,%s)",STR(self),STR(k));
}

tp_obj tp_get(TP,tp_obj self, tp_obj k) {
    int type = self.type;
    if (type == TP_DICT) {
        return _tp_dict_get(vm,self.dict.val,k,"tp_get");
    } else if (type == TP_LIST) {
        if (k.type == TP_NUMBER) {
            int l = tp_len(vm,self).number.val;
            int n = k.number.val;
            n = (n<0?l+n:n);
            return _tp_list_get(vm,self.list.val,n,"tp_get");
        } else if (k.type == TP_STRING) {
            if (strcmp("append",tp_cstring(k)) == 0) {
                return tp_method(vm,self,tp_append);
            } else if (strcmp("pop",tp_cstring(k)) == 0) {
                return tp_method(vm,self,tp_pop);
            } else if (strcmp("index",tp_cstring(k)) == 0) {
                return tp_method(vm,self,tp_index);
            } else if (strcmp("sort",tp_cstring(k)) == 0) {
                return tp_method(vm,self,tp_sort);
            } else if (strcmp("extend",tp_cstring(k)) == 0) {
                return tp_method(vm,self,tp_extend);
            }
        }
    } else if (type == TP_STRING) {
        if (k.type == TP_NUMBER) {
            int l = self.string.len;
            int n = k.number.val;
            n = (n<0?l+n:n);
            if (n < 0 || n >= l) {
                tp_raise(None,"tp_get(%s,%s)",STR(self),STR(k));
            }
            return tp_printf(vm,"%c",self.string.val[n]);
        } else if (k.type == TP_STRING) {
            if (strcmp("join",tp_cstring(k)) == 0) {
                return tp_method(vm,self,tp_join);
            } else if (strcmp("split",tp_cstring(k)) == 0) {
                return tp_method(vm,self,tp_split);
            } else if (strcmp("index",tp_cstring(k)) == 0) {
                return tp_method(vm,self,tp_str_index);
            } else if (strcmp("strip",tp_cstring(k)) == 0) {
                return tp_method(vm,self,tp_strip);
            } else if (strcmp("replace",tp_cstring(k)) == 0) {
                return tp_method(vm,self,tp_replace);
            }
        }
    } else if (type == TP_PARAMS) {
        if (k.type == TP_STRING) {
            if (k.string.val[0] == '[') {
                tp_obj r = _tp_list_copy(vm,self);
                self.list.val->len = 0;
                return r;
            } else if (k.string.val[0] == '{') {
                tp_obj r = _tp_dict_copy(vm,self);
                // should empty our dict here ..
                return r;
            } else if (self.list.val->len) {
                return _tp_list_pop(vm,self.list.val,0,"tp_get");
            } else {
                return _tp_dict_get(vm,self.dict.val,k,"tp_get");
            }
        }
    } else if (type == TP_DATA) {
        return self.data.meta->get(vm,self,k);
    }

    if (k.type == TP_LIST) {
        int a,b,l,i;
        tp_obj tmp;
        l = tp_len(vm,self).number.val;
        tmp = tp_get(vm,k,tp_number(0));
        if (tmp.type == TP_NUMBER) { a = tmp.number.val; }
        else if(tmp.type == TP_NONE) { a = 0; }
        else { tp_raise(None,"%s is not a number",STR(tmp)); }
        tmp = tp_get(vm,k,tp_number(1));
        if (tmp.type == TP_NUMBER) { b = tmp.number.val; }
        else if(tmp.type == TP_NONE) { b = l; }
        else { tp_raise(None,"%s is not a number",STR(tmp)); }
        a = (a<0?l+a:a); b = (b<0?l+b:b);
        if (type == TP_LIST) {
            tp_obj r = tp_list(vm,0,0);
            for (i=a; i<b; i++) {
                _tp_list_append(vm,r.list.val,tp_get(vm,self,tp_number(i)));
            }
            return r;
        } else if (type == TP_STRING) {
            char *ptr = tp_malloc(b-a+1);
            memcpy(ptr,self.string.val+a,b-a); ptr[b-a]=0;
            return tp_string_t(vm,ptr,b-a);
        }
    }



    tp_raise(None,"tp_get(%s,%s)",STR(self),STR(k));
}

tp_obj tp_dget(TP,tp_obj self, tp_obj k, tp_obj d) {
    if (tp_has(vm,self,k).number.val) { return tp_get(vm,self,k); }
    return d;
}
void tp_iget(TP,tp_obj *r, tp_obj self, tp_obj k) {
    if (tp_has(vm,self,k).number.val) { *r = tp_get(vm,self,k); tp_grey(vm,*r); }
}

void tp_set(TP,tp_obj self, tp_obj k, tp_obj v) {
    int type = self.type;

    if (type == TP_DICT) {
        _tp_dict_set(vm,self.dict.val,k,v);
        return;
    } else if (type == TP_LIST) {
        _tp_list_set(vm,self.list.val,k.number.val,v,"tp_set");
        return;
    } else if (type == TP_PARAMS && k.type == TP_STRING) {
        if (k.string.val[0] == '[') {
            tp_extend(tp_params(vm,2,(tp_obj[]){self,v}));
        } else if (k.string.val[0] == '{') {
            tp_merge(tp_params(vm,2,(tp_obj[]){self,v}));
        } else {
            _tp_dict_set(vm,self.dict.val,k,v);
        }
        return;
    } else if (type == TP_DATA) {
        self.data.meta->set(vm,self,k,v);
        return;
    }
    tp_raise(,"tp_set(%s,%s,%s)",STR(self),STR(k),STR(v));
}

tp_obj tp_add(TP,tp_obj a, tp_obj b) {
    if (a.type == TP_NUMBER && a.type == b.type) {
        return tp_number(a.number.val+b.number.val);
    } else if (a.type == TP_STRING && a.type == b.type) {
        int al = a.string.len, bl = b.string.len;
        char *s = tp_malloc(al+bl+1); s[al+bl] = 0;
        memcpy(s,a.string.val,al); memcpy(s+al,b.string.val,bl);
        return tp_string_t(vm,s,al+bl);
    } else if (a.type == TP_LIST && a.type == b.type) {
        tp_obj r = tp_copy(tp_params(vm,1,(tp_obj[]){a}));
        tp_extend(tp_params(vm,2,(tp_obj[]){r,b}));
        return r;
    }
    tp_raise(None,"tp_add(%s,%s)",STR(a),STR(b));
}


tp_obj tp_mul(TP,tp_obj a, tp_obj b) {
    if (a.type == TP_NUMBER && a.type == b.type) {
        return tp_number(a.number.val*b.number.val);
    } else if (a.type == TP_STRING && b.type == TP_NUMBER) {
        int al = a.string.len; int n = b.number.val;
        char *s = tp_malloc(al*n+1); s[al*n] = 0;
        int i; for (i=0; i<n; i++) { memcpy(s+al*i,a.string.val,al); }
        return tp_string_t(vm,s,al*n);
    }
    tp_raise(None,"tp_mul(%s,%s)",STR(a),STR(b));
}


tp_obj tp_len(TP,tp_obj self) {
    int type = self.type;
    if (type == TP_STRING) {
        return tp_number(self.string.len);
    } else if (type == TP_DICT) {
        return tp_number(self.dict.val->len);
    } else if (type == TP_LIST) {
        return tp_number(self.list.val->len);
    }
    tp_raise(None,"tp_len(%s)",STR(self));
}

tp_obj tp_cmp(TP,tp_obj a, tp_obj b) {
    int type = a.type;
    if (type != b.type) {
        return tp_number(a.type-b.type);
    } else if (type == TP_NONE) {
        return tp_number(0);
    } else if (type == TP_NUMBER) {
        return tp_number(a.number.val-b.number.val);
    } else if (type == TP_STRING) {
        int v = memcmp(a.string.val,b.string.val,_tp_min(a.string.len,b.string.len));
        if (v == 0) { v = a.string.len-b.string.len; }
        return tp_number(v);
    } else if (type == TP_LIST) {
        return tp_number(a.list.val - b.list.val);
    } else if (type == TP_DICT) {
        return tp_number(a.dict.val - b.dict.val);
    } else if (type == TP_PARAMS) {
        return tp_number(a.list.val - b.list.val);
    } else if (type == TP_DEF || type == TP_FNC) {
        if (a.fnc.val != b.fnc.val) { return tp_number(a.fnc.val-b.fnc.val); }
        return tp_number(a.fnc.self-b.fnc.self);
    }
    tp_raise(tp_number(0),"tp_cmp(%s,%s)",STR(a),STR(b));
}

#define TP_OP(name,expr) \
    tp_obj name(TP,tp_obj _a,tp_obj _b) { \
    if (_a.type == TP_NUMBER && _a.type == _b.type) { \
        float a = _a.number.val; float b = _b.number.val; \
        return tp_number(expr); \
    } \
    tp_raise(None,"%s(%s,%s)",#name,STR(_a),STR(_b)); \
}

TP_OP(tp_and,((int)a)&((int)b));
TP_OP(tp_or,((int)a)|((int)b));
TP_OP(tp_mod,((int)a)%((int)b));
TP_OP(tp_lsh,((int)a)<<((int)b));
TP_OP(tp_rsh,((int)a)>>((int)b));
TP_OP(tp_sub,a-b);
TP_OP(tp_div,a/b);
TP_OP(tp_pow,pow(a,b));

#ifndef TP_COMPILER
#define TP_COMPILER 1
#endif

tp_obj tp_import(TP,char *,char *,void *);
tp_obj tp_call(TP,char*,char*,tp_obj);

#if TP_COMPILER
#include "tpbc.c"
void tp_compiler(tp_vm *vm) {
    tp_import(vm,0,"pylang",tp_pylang);
    tp_import(vm,0,"dumbparse",tp_dumbparse);
    tp_import(vm,0,"dump2vm",tp_dump2vm);
    tp_import(vm,0,"py2c",tp_py2c);
    tp_call(vm,"py2c","_init",None);
}
#else
void tp_compiler(tp_vm *vm) {
}
#endif

#include "vm.c"

//
