#ifndef TP_H
#define TP_H

#include <setjmp.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <stdint.h>

#define tp_malloc(x) calloc((x),1)
#define tp_realloc(x,y) realloc(x,y)
#define tp_free(x) free(x);

union tp_obj;
typedef enum tp_types {
    TP_NONE,TP_NUMBER,TP_STRING,TP_DICT,
    TP_LIST,TP_PARAMS,TP_FNC,TP_DEF,TP_DATA,
} tp_types;

typedef struct tp_number_ {
    int type;
    float val;
} tp_number_;

typedef struct tp_string_ {
    int type;
    char *val;
    int len;
    int isptr;
} tp_string_;

typedef struct _tp_list {
    union tp_obj *items;
    int len;
    int alloc;
} _tp_list;

typedef struct tp_list_ {
    int type;
    _tp_list *val;
    void *_;
} tp_list_;
typedef struct tp_dict_ {
    int type;
    void *_;
    struct _tp_dict *val;
} tp_dict_;
typedef struct tp_params_ {
    int type;
    _tp_list *list;
    struct _tp_dict *dict;
    struct tp_vm *vm;
} tp_params_;


typedef struct tp_fnc_ {
    int type;
    void *val;
    void *self;
    void *globals;
} tp_fnc_;

typedef struct tp_data_ {
    int type;
    void *val;
    struct tp_meta *meta;
} tp_data_;

typedef union tp_obj {
    int type;
    tp_number_ number;
    tp_string_ string;
    tp_dict_ dict;
    tp_list_ list;
    tp_params_ params;
    tp_fnc_ fnc;
    tp_fnc_ def;
    tp_data_ data;
} tp_obj;


typedef struct tp_item {
    int used;
    int hash;
    tp_obj key;
    tp_obj val;
} tp_item;
typedef struct _tp_dict {
    tp_item *items;
    int len;
    int alloc;
    int cur;
    int mask;
    int used;
} _tp_dict;

typedef struct tp_code_r { unsigned char i,a,b,c; } tp_code_r;
typedef struct tp_code_s { char val[4]; } tp_code_s;
typedef struct tp_code_n { float val; } tp_code_n;
typedef union tp_code {
    unsigned char i;
    tp_code_r regs;
    tp_code_s string;
    tp_code_n number;
} tp_code;

typedef struct tp_frame_ {
    tp_code *codes;
    int cur;
    int jmp;
    tp_obj *regs;
    tp_obj *ret_dest;
    tp_obj fname;
    tp_obj name;
    tp_obj line;
    tp_obj globals;
    int lineno;
} tp_frame_;

#define TP_GCMAX 4096
typedef struct tp_gc {
    _tp_dict *white;
    _tp_list *grey;
    _tp_list *black;
    _tp_dict *strings;
    int target;
    int steps;
} tp_gc;


#define TP_FRAMES 64
#define TP_REGS_PER_FRAME 256
#define TP_REGS (TP_REGS_PER_FRAME*TP_FRAMES)
typedef struct tp_vm {
    tp_obj builtins;
    tp_obj modules;
    tp_frame_ frames[TP_FRAMES];
    tp_obj params;
    tp_obj _regs;
    tp_obj *regs;
    tp_obj root;
    jmp_buf buf;
    int jmp;
    tp_obj ex;
    int cur;
    tp_gc *gc;
} tp_vm;

#define TP tp_vm *vm
typedef struct tp_meta {
    tp_obj (*get)(TP,tp_obj,tp_obj);
    tp_obj (*set)(TP,tp_obj,tp_obj,tp_obj);
    tp_obj (*free)(TP,tp_obj);
//     tp_obj (*del)(TP,tp_obj,tp_obj);
//     tp_obj (*has)(TP,tp_obj,tp_obj);
//     tp_obj (*len)(TP,tp_obj);
} tp_meta;

// NOTE: these are the few out of namespace items for convenience
#define None ((tp_obj){TP_NONE})
#define True tp_number(1)
#define False tp_number(0)
#define STR(v) (tp_cstring(tp_str(vm,(v))))

tp_obj tp_string(char *val);
tp_obj tp_string_n(char *val,int );
void tp_set(TP,tp_obj,tp_obj,tp_obj);
tp_obj tp_get(TP,tp_obj,tp_obj);
tp_obj tp_len(TP,tp_obj);
tp_obj tp_str(TP,tp_obj);
tp_obj *tp_ptr(tp_obj o);
tp_obj tp_cmp(TP,tp_obj,tp_obj);
tp_obj tp_is(TP,tp_obj,tp_obj);
tp_obj tp_dget(TP,tp_obj,tp_obj,tp_obj);
void _tp_raise(TP,tp_obj);
// char *tp_printf(char *fmt,...);
tp_obj tp_printf(TP,char *fmt,...);
char *tp_cstring(tp_obj);
void tp_track(TP,tp_obj);
tp_obj tp_track_s(TP,tp_obj);
void tp_grey(TP,tp_obj);
void tp_follow(TP,tp_obj);
void tp_delete(TP,tp_obj v);
tp_obj tp_string_t(TP, char *v, int n);

#define tp_raise(r,fmt,...) { \
    _tp_raise(vm,tp_printf(vm,fmt,__VA_ARGS__)); \
    return r; \
}
#define TP_PBEGIN(params) \
    tp_obj __params = (params); \
    tp_vm *vm = __params.params.vm; \
    vm += 0;
#define TP_POBJ(k) (tp_get(vm,__params,tp_string(k)))
inline tp_obj _tp_check(TP,tp_obj v, int t) {
    if (v.type != t) { tp_raise(None,"_tp_check(%s,%d)",STR(v),t); }
    return v;
}
#define TP_PTYPE(k,t) _tp_check(vm,TP_POBJ(k),t)
#define TP_PNEXT() _tp_list_pop(vm,__params.list.val,0,"TP_POP")
#define TP_PNUMBER(k) (TP_PTYPE(k,TP_NUMBER).number.val)
#define TP_PSTRING(k) (tp_cstring(TP_PTYPE(k,TP_STRING)))
#define TP_PDEFAULT(k,d) (tp_dget(vm,__params,tp_string(k),d))
#define TP_PEND \
    __params.list.val->len = 0; \
    (__params.type = TP_NONE)
#define TP_PLOOP(e) \
    int __l = __params.list.val->len; \
    int __i; for (__i=0; __i<__l; __i++) { \
    (e) = _tp_list_get(vm,__params.list.val,__i,"TP_LOOP");
#define TP_PEOL \
    }

inline int _tp_min(int a, int b) { return (a<b?a:b); }
inline int _tp_max(int a, int b) { return (a>b?a:b); }
inline tp_obj tp_number(float v) { return (tp_obj)(tp_number_){TP_NUMBER,v}; }

#endif
//
