#include "vmsdl.c"

void tp_handle(TP);
tp_obj tp_exec(TP,tp_obj,tp_obj);


tp_vm *_tp_init(void) {
    tp_vm * vm = tp_malloc(sizeof(tp_vm));
    vm->cur = -1;
    vm->jmp = 0;
    vm->builtins = tp_dict_new();
    vm->modules = tp_dict_new();
    vm->params = tp_list_new(0);
    vm->ex = None;
    vm->_regs = tp_list_new(0);
    vm->root = tp_list_new(0);
    tp_gc_init(vm);
    _tp_list_append(vm,vm->root.list.val,vm->builtins);
    _tp_list_append(vm,vm->root.list.val,vm->modules);
    _tp_list_append(vm,vm->root.list.val,vm->_regs);
    _tp_list_append(vm,vm->root.list.val,vm->params);
    
    tp_track(vm,vm->builtins);
    tp_track(vm,vm->modules);
    tp_track(vm,vm->_regs);
    tp_track(vm,vm->params);
    
    tp_set(vm,vm->builtins,tp_string("MODULES"),vm->modules);
    tp_set(vm,vm->modules,tp_string("BUILTINS"),vm->builtins);
    tp_set(vm,vm->builtins,tp_string("BUILTINS"),vm->builtins);
    int i;
    for (i=0; i<TP_FRAMES; i++) {
        _tp_list_append(vm,vm->params.list.val,tp_params(vm,0,0));
    }
    for (i=0; i<TP_REGS; i++) {
        _tp_list_append(vm,vm->_regs.list.val,None);
    }
    vm->regs = vm->_regs.list.val->items;
    tp_full(vm);
    return vm;
}

void tp_deinit(TP) {
    while (vm->root.list.val->len) {
        _tp_list_pop(vm,vm->root.list.val,0,"tp_deinit");
    }
    tp_full(vm);
    tp_full(vm);
    tp_delete(vm,vm->root);
    tp_gc_deinit(vm);
    tp_free(vm);
}


tp_frame_ tp_frame(tp_obj globals,tp_code *codes,tp_obj *regs,tp_obj *ret_dest) {
    tp_frame_ f;
    f.globals = globals;
    f.codes = codes;
    f.cur = 0;
    f.jmp = 0;
    f.regs = regs;
    f.ret_dest = ret_dest;
    f.lineno = 0;
    f.line = tp_string("");
    f.name = tp_string("?");
    f.fname = tp_string("?");
    return f;
}

typedef enum tp_instructions {
    TP_IEOF,TP_IADD,TP_ISUB,TP_IMUL,TP_IDIV,TP_IPOW,TP_IAND,TP_IOR,TP_ICMP,TP_IGET,TP_ISET,
    TP_INUMBER,TP_ISTRING,TP_IGGET,TP_IGSET,TP_IMOVE,TP_IDEF,TP_IPASS,TP_IJUMP,TP_ICALL,
    TP_IRETURN,TP_IIF,TP_IDEBUG,TP_IEQ,TP_ILE,TP_ILT,TP_IDICT,TP_ILIST,TP_INONE,TP_ILEN,
    TP_ILINE,TP_IPARAMS,TP_IIGET,TP_IFILE,TP_INAME,TP_INE,TP_IHAS,TP_IRAISE,TP_ISETJMP,
    TP_IMOD,TP_ILSH,TP_IRSH,TP_IITER,TP_IDEL,
    TP_ITOTAL
} tp_instructions;

// char *tp_strings[TP_ITOTAL] = {
//     "EOF","ADD","SUB","MUL","DIV","POW","AND","OR","CMP","GET","SET","NUM",
//     "STR","GGET","GSET","MOVE","DEF","PASS","JUMP","CALL","RETURN","IF","DEBUG",
//     "EQ","LE","LT","DICT","LIST","NONE","LEN","LINE","PARAMS","IGET","FILE",
//     "NAME","NE","HAS","RAISE","SETJMP","MOD","LSH","RSH","ITER","DEL",
// };

#define VA ((int)e.regs.a)
#define VB ((int)e.regs.b)
#define VC ((int)e.regs.c)
#define RA regs[e.regs.a]
#define RB regs[e.regs.b]
#define RC regs[e.regs.c]
#define R0 regs[0]
#define NA regs[e.regs.a].number.val
#define NB regs[e.regs.b].number.val
#define NC regs[e.regs.c].number.val
#define NUM(n) tp_number(n)
#define UVBC (unsigned short)(((VB<<8)+VC))
#define SVBC (short)(((VB<<8)+VC))

#define GA tp_grey(vm,RA)

void _tp_raise(TP,tp_obj e) {
    if (!vm || !vm->jmp) {
        fprintf(stderr,"%s\n",STR(e));
        exit(-1);
        return;
    }
    vm->ex = e;
    tp_grey(vm,e);
    longjmp(vm->buf,1);
}

void tp_print_stack(TP) {
    printf("\n");
    int i;
    for (i=0; i<=vm->cur; i++) {
        printf("File \"%s\", line %d, in %s\n  %s\n",
            STR(vm->frames[i].fname),vm->frames[i].lineno,STR(vm->frames[i].name),STR(vm->frames[i].line));
    }
    printf("\nException:\n");
    printf("%s\n",tp_cstring(tp_str(vm,vm->ex)));
}



void tp_handle(TP) {
    int i;
    for (i=vm->cur; i>=0; i--) {
        if (vm->frames[i].jmp) { break; }
    }
    if (i >= 0) {
        vm->cur = i;
        vm->frames[i].cur = vm->frames[i].jmp;
        vm->frames[i].jmp = 0;
        return;
    }
    tp_print_stack(vm);
    exit(-1);
}

void _tp_call(TP,tp_obj *dest, tp_obj fnc, tp_obj params) {
    if (fnc.type == TP_FNC) {
        *dest = _tp_tcall(fnc,params);
        tp_grey(vm,*dest);
        return;
    }
    if (fnc.type == TP_DICT) {
        return _tp_call(vm,dest,tp_get(vm,fnc,tp_string("__call__")),params);
    }
    if (fnc.type == TP_DEF) {
        vm->cur += 1;
        vm->frames[vm->cur] = tp_frame(*(tp_obj*)fnc.def.globals,fnc.def.val,&vm->regs[vm->cur*TP_REGS_PER_FRAME],dest);
        if (fnc.def.self) {
            vm->frames[vm->cur].regs[0] = params;
            _tp_list_insert(vm,params.list.val,0,*(tp_obj*)fnc.def.self);
        } else {
            vm->frames[vm->cur].regs[0] = params;
        }
        return;
    }
    tp_print(tp_params(vm,1,(tp_obj[]){fnc}));
    tp_raise(,"tp_call: %s is not callable",STR(fnc));
}

void tp_return(tp_vm *vm, tp_obj v) {
    tp_obj *dest = vm->frames[vm->cur].ret_dest;
    if (dest) { *dest = v; tp_grey(vm,v); }
    memset(vm->frames[vm->cur].regs,0,TP_REGS_PER_FRAME*sizeof(tp_obj));
    vm->cur -= 1;
}

int tp_step(TP) {
    tp_frame_ *f = &vm->frames[vm->cur];
    tp_code e = f->codes[f->cur];
    tp_obj *regs = f->regs;
    
//     tp_gcinc(vm);
//     fprintf(stderr,"%d.%d: %s %d %d %d\n",vm->cur,f->cur,tp_strings[e.i],VA,VB,VC);
//     int i; for(i=0;i<16;i++) { fprintf(stderr,"%d: %s\n",i,STR(regs[i])); }
    switch (e.i) {
        case TP_IEOF:
            if (vm->cur) { tp_return(vm,None); return 0; }
            return -1;
            break;
        case TP_IADD: RA = tp_add(vm,RB,RC); break;
        case TP_ISUB: RA = tp_sub(vm,RB,RC); break;
        case TP_IMUL: RA = tp_mul(vm,RB,RC); break;
        case TP_IDIV: RA = tp_div(vm,RB,RC); break;
        case TP_IPOW: RA = tp_pow(vm,RB,RC); break;
        case TP_IAND: RA = tp_and(vm,RB,RC); break;
        case TP_IOR:  RA = tp_or(vm,RB,RC); break;
        case TP_IMOD:  RA = tp_mod(vm,RB,RC); break;
        case TP_ILSH:  RA = tp_lsh(vm,RB,RC); break;
        case TP_IRSH:  RA = tp_rsh(vm,RB,RC); break;
        case TP_ICMP: RA = tp_cmp(vm,RB,RC); break;
        case TP_INE: RA = tp_number(tp_cmp(vm,RB,RC).number.val!=0); break;
        case TP_IEQ: RA = tp_number(tp_cmp(vm,RB,RC).number.val==0); break;
        case TP_ILE: RA = tp_number(tp_cmp(vm,RB,RC).number.val<=0); break;
        case TP_ILT: RA = tp_number(tp_cmp(vm,RB,RC).number.val<0); break;
        case TP_IPASS: break;
        case TP_IIF: if (RA.number.val) { f->cur += 1; } break;
        case TP_IGET: RA = tp_get(vm,RB,RC); GA; break;
        case TP_IITER: RA = tp_iter(vm,RB,RC); GA; break;
        case TP_IHAS: RA = tp_has(vm,RB,RC); break;
        case TP_IIGET: tp_iget(vm,&RA,RB,RC); break;
        case TP_ISET: tp_set(vm,RA,RB,RC); break;
        case TP_IDEL: tp_del(vm,RA,RB); break;
        case TP_IMOVE: RA = RB; break;
        case TP_INUMBER: RA = tp_number(f->codes[++f->cur].number.val); break;
        case TP_ISTRING:
            RA = tp_string_n(f->codes[f->cur+1].string.val,UVBC);
            f->cur += (UVBC/4)+1;
            break;
        case TP_IDICT: RA = tp_dict(vm,VC,&RB); break;
        case TP_ILIST: RA = tp_list(vm,VC,&RB); break;
        case TP_IPARAMS: {
            tp_obj r = vm->params.list.val->items[vm->cur];
            int i; for (i=0; i<VC; i++) {
                _tp_list_append(vm,r.list.val,(&RB)[i]);
                }
            RA = r;
            }
            break;
        case TP_ILEN: RA = tp_len(vm,RB); break;
        case TP_IJUMP: f->cur += SVBC; return 0; break;
        case TP_ISETJMP: f->jmp = f->cur+SVBC; break;
        case TP_ICALL: _tp_call(vm,&RA,RB,RC); break;
        case TP_IGGET:
            if (tp_has(vm,f->globals,RB).number.val) {
                RA = tp_get(vm,f->globals,RB);
            } else { RA = tp_get(vm,vm->builtins,RB); }
            GA;
            break;
        case TP_IGSET: tp_set(vm,f->globals,RA,RB); break;
        case TP_IDEF:
            RA = tp_def(vm,f->codes[f->cur+1].string.val,f->globals);
            f->cur += SVBC;
            return 0;
            break;
        case TP_IRETURN: tp_return(vm,RA); return 0; break;
        case TP_IRAISE: _tp_raise(vm,RA); return 0; break;
        case TP_IDEBUG:
            tp_print(tp_params(vm,2,(tp_obj[]){tp_string("DEBUG:"),RA}));
            break;
        case TP_INONE: RA = None; break;
        case TP_ILINE:
            f->line = tp_string_n(f->codes[f->cur+1].string.val,VA*4-1);
            f->cur += VA; f->lineno = UVBC;
            break;
        case TP_IFILE: f->fname = RA; break;
        case TP_INAME: f->name = RA; break;
        default: tp_raise(0,"tp_step: invalid instruction %d",e.i); break;
    }
    f->cur += 1;
    return 0;
}

void tp_run(TP,int cur) {
    if (vm->jmp) { tp_raise(,"tp_run(%d) called recusively",cur); }
    vm->jmp = 1; if (setjmp(vm->buf)) { tp_handle(vm); }
    while (vm->cur >= cur && tp_step(vm) != -1);
    vm->cur = cur-1; vm->jmp = 0;
}

tp_obj _tp_import(TP,char *fname, char *name, void *codes) {
    tp_obj code = None;
    if (!codes) {
        if (!tp_file_exists(fname)) {
            tp_raise(None,"_tp_import(%s) - file does not exist",fname);
        }
        FILE *f;
        long l = tp_FileSize(fname);
        f = fopen(fname,"rb");
        codes = tp_malloc(l);
        fread(codes,1,l,f);
        fclose(f);
        code = tp_string_t(vm,(char*)codes,l);
    } else {
        code.type = TP_DATA;
        code.data.val = codes;
    }

    vm->cur += 1;
    tp_obj g = tp_dict(vm,4,(tp_obj[]){tp_string("__name__"),tp_string(name),tp_string("CODE"),code});
    vm->frames[vm->cur] = tp_frame(g,codes,&vm->regs[vm->cur*TP_REGS_PER_FRAME],0);
    tp_set(vm,vm->modules,tp_string(name),g);
    
    if (!vm->jmp) { tp_run(vm,vm->cur); }
    
    return g;
}



tp_obj tp_exec_(tp_obj params) {
    TP_PBEGIN(params);
    tp_obj code = TP_POBJ("code");
    tp_obj globals = TP_POBJ("globals");
    TP_PEND;
    
    vm->cur += 1;
    vm->frames[vm->cur] = tp_frame(globals,(void*)code.string.val,&vm->regs[vm->cur*TP_REGS_PER_FRAME],0);
    return None;
}

tp_obj tp_call(TP, char *mod, char *fnc, tp_obj params) {
    tp_obj tmp;
    tp_obj r = None;
    tmp = tp_get(vm,vm->modules,tp_string(mod));
    tmp = tp_get(vm,tmp,tp_string(fnc));
    _tp_call(vm,&r,tmp,params);
    tp_run(vm,vm->cur);
    return r;
}

tp_obj tp_import_(tp_obj params) {
    TP_PBEGIN(params);
    tp_obj mod = TP_POBJ("module");
    TP_PEND;

    if (tp_has(vm,vm->modules,mod).number.val) {
        return tp_get(vm,vm->modules,mod);
    }

    char *s = tp_cstring(mod);
    tp_obj r = _tp_import(vm,tp_cstring(tp_add(vm,mod,tp_string(".tpc"))),s,0);
    return r;
}

#define TP_IB(a,b) tp_set(vm,ctx,tp_string(a),tp_fnc(b))

tp_obj tp_builtins(TP) {
    tp_obj ctx = vm->builtins;
    TP_IB("print",tp_print); TP_IB("range",tp_range); TP_IB("min",tp_min);
    TP_IB("max",tp_max); TP_IB("bind",tp_bind); TP_IB("copy",tp_copy);
    TP_IB("import",tp_import_); TP_IB("len",tp_len_); TP_IB("assert",tp_assert);
    TP_IB("str",tp_str2); TP_IB("float",tp_float); TP_IB("system",tp_system);
    TP_IB("istype",tp_istype); TP_IB("chr",tp_chr); TP_IB("save",tp_save);
    TP_IB("load",tp_load); TP_IB("fpack",tp_fpack); TP_IB("abs",tp_abs);
    TP_IB("int",tp_int); TP_IB("exec",tp_exec_); TP_IB("exists",tp_exists);
    TP_IB("mtime",tp_mtime); TP_IB("number",tp_float); TP_IB("round",tp_round);
    TP_IB("ord",tp_ord);
    sdl_init(vm);
    return ctx;
}


void tp_args(TP,int argc, char *argv[]) {
    tp_obj self = tp_list(vm,0,0);
    int n;
    for (n=1; n<argc; n++) {
        _tp_list_append(vm,self.list.val,tp_string(argv[n]));
    }
    tp_set(vm,vm->builtins,tp_string("ARGV"),self);
}

tp_obj _tp_import_smart(TP,char*fname,char*name,void*code) {
    if ((fname && strstr(fname,".tpc")) || code) {
        return _tp_import(vm,fname,name,code);
    }
    return tp_call(vm,"py2c","import_fname",tp_params(vm,2,(tp_obj[]){tp_string(fname),tp_string(name)}));
}
    

tp_obj tp_main(TP,char *fname, void *code) {
    return _tp_import_smart(vm,fname,"__main__",code);
}
tp_obj tp_import(TP, char *fname, char *name, void *code) {
    return _tp_import_smart(vm,fname,name,code);
}

tp_obj tp_compile(TP, tp_obj text, tp_obj fname) {
    return tp_call(vm,"BUILTINS","compile",tp_params(vm,2,(tp_obj[]){text,fname}));
}

tp_obj tp_exec(TP,tp_obj code, tp_obj globals) {
    tp_obj r=None;
    vm->cur += 1;
    vm->frames[vm->cur] = tp_frame(globals,(void*)code.string.val,&vm->regs[vm->cur*TP_REGS_PER_FRAME],&r);
    tp_run(vm,vm->cur);
    return r;
}

tp_obj tp_eval(TP, char *text, tp_obj globals) {
    tp_obj code = tp_compile(vm,tp_string(text),tp_string("<eval>"));
    return tp_exec(vm,code,globals);
}

tp_vm *tp_init(int argc, char *argv[]) {
    tp_vm *vm = _tp_init();
    tp_builtins(vm);
    tp_args(vm,argc,argv);
    tp_compiler(vm);
    return vm;
}
    

//
